///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: _epb.cpp,v 1.4 2001/05/22 22:26:08 epocman Exp $
//	Subject : IBPP, internal EPB class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* EPB == Event Parameter Block, see Interbase jrd/alt.c and 6.0 C-API
//	* Tabulations should be set every four characters when editing this file.
//
//	SPECIAL WARNING COMMENT (by Olivier Mascia, 2000 Nov 12)
//	This Event Parameter Block handling is not publicly documented, per se, in
//	the ibase.h header file or in the IB 6.0 documentation. This documentation
//	suggests to use the API isc_event_block to construct such EPBs.
//	Unfortunately, this API takes a variable number of parameters to specify
//	the list of event names. In addition, the documentation warn on not using
//	more than 15 names. This is a sad limitation, partly because the maximum
//	number of parameters safely processed in such an API is very compiler
//	dependant and also because isc_event_counts() is designed to return counts
//	through the IB status vector which is a vector of 20 longs !
//	From reverse engineering of the isc_event_block() API in
//	source file jrd/alt.c (as available on fourceforge.net/project/InterBase as
//	of 2000 Nov 12), it looks like the internal format of those EPB is simple.
//	An EPB starts by a byte with value 1. A version identifier of some sort.
//	Then a small cluster is used for each event name. The cluster starts with
//	a byte for the length of the event name (no final '\0'). Followed by the N
//	characters of the name itself (no final '\0'). The cluster ends with 4 bytes
//	preset to 0.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

static const int BUFFERINCR = 128;
static const int ARRAYINCR = 8;

void _EPB::Define(const char* eventname, IBPP::EventInterface* objref)
{
	int len = strlen(eventname);
	if (len == 0) throw ExceptionImpl("_EPB::Define",
			"Zero length event names not permitted");
	if (len > 64) throw ExceptionImpl("_EPB::Define",
			"IBPP Max Event Name length is 64 characters");

	// 1) Grow the events/result buffers, if needed
	if (_BufSize + len + 5 > _BufAlloc)
	{
		// The buffers are too small. Let's re-allocate them.
		int newBufAlloc = _BufAlloc + BUFFERINCR;
		char* newEvnBuffer = new char [newBufAlloc];
		char* newResBuffer = new char [newBufAlloc];
		if (newEvnBuffer == NULL || newResBuffer == NULL)
			throw ExceptionImpl("_EPB::Define", "Memory allocation error");
		memcpy(newEvnBuffer, _EvnBuffer, _BufSize);
		memcpy(newResBuffer, _ResBuffer, _BufSize);
		delete [] _EvnBuffer;
		delete [] _ResBuffer;
		_EvnBuffer = newEvnBuffer;
		_ResBuffer = newResBuffer;
		_BufAlloc = newBufAlloc;
	}

	// 2) Update the events/result buffers
	char* p = _EvnBuffer + _BufSize;

	*p++ = (char)len;
	for (int i = 0; i < len; i++) *p++ = eventname[i];
	*p++ = (char)1;		// We initialize the counts to 1
	*p++ = (char)0;
	*p++ = (char)0;
	*p = (char)0;

	_BufSize += (len + 5);

	// 3) Grow the objref array, if needed
	if (_EvnCount == _EvnAlloc)
	{
		// The array is too small. Let's re-allocate it.
		int newEvnAlloc = _EvnAlloc + ARRAYINCR;
		IBPP::EventInterface** newObjRefs = new IBPP::EventInterface*[newEvnAlloc];
		if (newObjRefs == NULL)
			throw ExceptionImpl("_EPB::Define", "Memory allocation error");
		memcpy(newObjRefs, _ObjRefs, _EvnCount);
		delete [] _ObjRefs;
		_ObjRefs = newObjRefs;
		_EvnAlloc = newEvnAlloc;
	}

	// 4) Update the objref array
	_ObjRefs[_EvnCount] = objref;
	_EvnCount++;
}

void _EPB::Drop(const char* eventname)
{
	if (eventname == NULL) throw ExceptionImpl("_EPB::Drop",
			"NULL pointer reference detected.");

	// 1) Find the event by its name, throw if not found

	// 2) Drop this event by shifting the buffers and array contents down

	throw ExceptionImpl("_EPB::Drop",
		"Not yet implemented, work in progress, contact om@tipgroup.com");
}

void _EPB::FireActions(IBPP::Database* db)
{
	int i;
	int j = 1;
	unsigned int * pold;
	unsigned int * pnew;

	for (i = 0; i < _EvnCount; i++)
	{
		// Saves a local copy of the event name
		char name[65];
		memcpy(name, &_EvnBuffer[j+1], _EvnBuffer[j]);
		name[_EvnBuffer[j]] = '\0';

		// Skips in the buffer to the count value
		j = j + _EvnBuffer[j] + 1;	// Add [size] bytes + 1 for the size
		pold = (unsigned int *) &_EvnBuffer[j];
		pnew = (unsigned int *) &_ResBuffer[j];
		if (*pnew > *pold)
		{
			// Fire the action
			try { _ObjRefs[i]->ibppEventHandler(db, name, *pnew - *pold); }
				catch (...) { }
		}
		j += 4;
	}

	// Copy over the results buffer to the events buffer to prepare for
	// next asynchronous wait.
	memcpy(_EvnBuffer, _ResBuffer, _BufSize);
}

_EPB::_EPB()
{
	_BufAlloc = BUFFERINCR;	// Let's start small, will grow buffers as needed
	_EvnAlloc = ARRAYINCR;	// Let's start small, will grow arrays as needed
	_EvnCount = 0;		// No events yet defined

	// Allocate the buffers
	_EvnBuffer = new char [_BufAlloc];
	_ResBuffer = new char [_BufAlloc];
	if (_EvnBuffer == NULL || _ResBuffer == NULL)
		throw ExceptionImpl("_EPB::_EPB", "Memory allocation error");

	// Initialize the events buffer
	_EvnBuffer[0] = (char)1;	// First byte is a 'one'. Documentation ??
	_BufSize = 1;

	// Allocate small arrays for objrefs and methods addresses
	_ObjRefs = new IBPP::EventInterface*[_EvnAlloc];
	if (_ObjRefs == NULL)
		throw ExceptionImpl("_EPB::_EPB", "Memory allocation error");
}

_EPB::~_EPB()
{
	if (_BufAlloc != 0)
	{
		delete [] _EvnBuffer;
		delete [] _ResBuffer;
		_EvnBuffer = _ResBuffer = NULL;
		_BufSize = _BufAlloc = 0;
	}

	if (_EvnAlloc != 0)
	{
		delete [] _ObjRefs;
		_ObjRefs = NULL;
		_EvnCount = _EvnAlloc = 0;
	}
}

//
//	EOF
//
