///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: blob.cpp,v 1.6 2001/11/03 21:01:43 epocman Exp $
//	Subject : IBPP, Blob class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef IBPP_WINDOWS
#include <windows.h>
#endif

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <stdio.h>

using namespace ibpp_internals;

//	(((((((( OBJECT INTERFACE IMPLEMENTATION ))))))))

void BlobImpl::AttachDatabase(IBPP::Database* database)
{
	if (database == NULL) throw ExceptionImpl("Blob::AttachDatabase",
			"Can't attach a NULL Database object.");

	if (_Database != NULL) _Database->DetachBlob(this);
	_Database = dynamic_cast<DatabaseImpl*>(database);
	_Database->AttachBlob(this);
}

void BlobImpl::AttachTransaction(IBPP::Transaction* transaction)
{
	if (transaction == NULL) throw ExceptionImpl("Blob::AttachTransaction",
			"Can't attach a NULL Transaction object.");

	if (_Transaction != NULL) _Transaction->DetachBlob(this);
	_Transaction = dynamic_cast<TransactionImpl*>(transaction);
	_Transaction->AttachBlob(this);
}

void BlobImpl::DetachDatabase(void)
{
	if (_Database == NULL) throw ExceptionImpl("Blob::DetachDatabase",
			"No Database was attached.");

	_Database->DetachBlob(this);
	_Database = NULL;
}

void BlobImpl::DetachTransaction(void)
{
	if (_Transaction == NULL) throw ExceptionImpl("Blob::DetachTransaction",
			"No Transaction was attached.");

	_Transaction->DetachBlob(this);
	_Transaction = NULL;
}

IBPP::Database* BlobImpl::GetDatabase(void)
{
	if (_Database == NULL) throw ExceptionImpl("Blob::GetDatabase",
			"No Database is attached.");
	return _Database;
}

IBPP::Transaction* BlobImpl::GetTransaction(void)
{
	if (_Transaction == NULL) throw ExceptionImpl("Blob::GetTransaction",
			"No Transaction is attached.");
	return _Transaction;
}

void BlobImpl::Open(void)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Blob::Open", "Blob already opened.");
	if (_Database == NULL)
		throw ExceptionImpl("Blob::Open", "No Database is attached.");
	if (_Transaction == NULL)
		throw ExceptionImpl("Blob::Open", "No Transaction is attached.");
	if (! _IdAssigned)
		throw ExceptionImpl("Blob::Open", "Blob Id is not assigned.");

	_IBS status;
	(*GDS::gds()->_open_blob2)(status.Self(), _Database->GetHandlePtr(),
		_Transaction->GetHandlePtr(), &_Handle, &_Id, 0, NULL);
	if (status.Errors())
		throw ExceptionImpl(&status, "Blob::Open", "isc_open_blob2 failed.");
	_WriteMode = false;
}

void BlobImpl::Create(void)
{
	if (_Handle != NULL)
		throw ExceptionImpl("Blob::Create", "Blob already opened.");
	if (_Database == NULL)
		throw ExceptionImpl("Blob::Create", "No Database is attached.");
	if (_Transaction == NULL)
		throw ExceptionImpl("Blob::Create", "No Transaction is attached.");

	_IBS status;
	(*GDS::gds()->_create_blob)(status.Self(), _Database->GetHandlePtr(),
		_Transaction->GetHandlePtr(), &_Handle, &_Id);
	if (status.Errors())
		throw ExceptionImpl(&status, "Blob::Create",
			"isc_create_blob failed.");
	_IdAssigned = true;
	_WriteMode = true;
}

void BlobImpl::Close(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Blob::Close", "The Blob is not opened");

	_IBS status;
	(*GDS::gds()->_close_blob)(status.Self(), &_Handle);
	if (status.Errors())
		throw ExceptionImpl(&status, "Blob::Close", "isc_close_blob failed.");
	_Handle = NULL;
}

void BlobImpl::Cancel(void)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Blob::Cancel", "The Blob is not opened");
	if (! _WriteMode)
		throw ExceptionImpl("Blob::Cancel", "Can't cancel a Blob opened for read");

	_IBS status;
	(*GDS::gds()->_cancel_blob)(status.Self(), &_Handle);
	if (status.Errors())
		throw ExceptionImpl(&status, "Blob::Cancel", "isc_cancel_blob failed.");
	_Handle = NULL;
	_IdAssigned = false;
}

int BlobImpl::Read(void* buffer, int size)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Blob::Read", "The Blob is not opened");
	if (_WriteMode)
		throw ExceptionImpl("Blob::Read", "Can't read from Blob opened for write");
	if (size < 1 || size > (64*1024-1))
		throw ExceptionImpl("Blob::Read", "Invalid segment size (max 64Kb-1)");

	_IBS status;
	unsigned short bytesread;
	int result = (*GDS::gds()->_get_segment)(status.Self(), &_Handle, &bytesread,
					(unsigned short)size, (char*)buffer);
	if (result == isc_segstr_eof) return 0;	// Fin du blob
	if (result != isc_segment && status.Errors())
		throw ExceptionImpl(&status, "Blob::Read", "isc_get_segment failed.");
	return (int)bytesread;
}

void BlobImpl::Write(void* buffer, int size)
{
	if (_Handle == NULL)
		throw ExceptionImpl("Blob::Write", "The Blob is not opened");
	if (! _WriteMode)
		throw ExceptionImpl("Blob::Write", "Can't write to Blob opened for read");
	if (size < 1 || size > (64*1024-1))
		throw ExceptionImpl("Blob::Write", "Invalid segment size (max 64Kb-1)");

	_IBS status;
	(*GDS::gds()->_put_segment)(status.Self(), &_Handle,
		(unsigned short)size, (char*)buffer);
	if (status.Errors())
		throw ExceptionImpl(&status, "Blob::Write", "isc_put_segment failed.");
}

void BlobImpl::GetInfo(int* Size, int* Largest, int* Segments)
{
	char items[] = {isc_info_blob_total_length,
					isc_info_blob_max_segment,
					isc_info_blob_num_segments};

	if (_Handle == NULL)
		throw ExceptionImpl("Blob::GetInfo", "The Blob is not opened");

	_IBS status;
	_RB result(100);
	(*GDS::gds()->_blob_info)(status.Self(), &_Handle, sizeof(items), items,
		result.Size(), result.Self());
	if (status.Errors())
		throw ExceptionImpl(&status, "Blob::GetInfo", "isc_blob_info failed.");

	if (Size != NULL) *Size = result.GetValue(isc_info_blob_total_length);
	if (Largest != NULL) *Largest = result.GetValue(isc_info_blob_max_segment);
	if (Segments != NULL) *Segments = result.GetValue(isc_info_blob_num_segments);
}

void BlobImpl::Release(IBPP::Blob*& Self)
{
	if (this != dynamic_cast<BlobImpl*>(Self))
		throw ExceptionImpl("Blob::Release", "Invalid Release()");
	delete this;
	Self = NULL;
}

//	(((((((( OBJECT INTERNAL METHODS ))))))))

void BlobImpl::Init(void)
{
	_IdAssigned = false;
	_WriteMode = false;
	_Handle = NULL;
	_Database = NULL;
	_Transaction = NULL;
}

void BlobImpl::SetId(ISC_QUAD* quad)
{
	if (_Handle != NULL)
		throw ExceptionImpl("BlobImpl::SetId", "Can't set Id on an opened BlobImpl.");
	if (quad == NULL)
		throw ExceptionImpl("BlobImpl::SetId", "NULL Id reference detected.");

	memcpy(&_Id, quad, sizeof(_Id));
	_IdAssigned = true;
}

void BlobImpl::GetId(ISC_QUAD* quad)
{
	if (_Handle != NULL)
		throw ExceptionImpl("BlobImpl::GetId", "Can't get Id on an opened BlobImpl.");
	if (! _WriteMode)
		throw ExceptionImpl("BlobImpl::GetId", "Can only get Id of a newly created Blob.");
	if (quad == NULL)
		throw ExceptionImpl("BlobImpl::GetId", "NULL Id reference detected.");

	memcpy(quad, &_Id, sizeof(_Id));
}

BlobImpl::BlobImpl(DatabaseImpl* database, TransactionImpl* transaction)
{
	Init();
	AttachDatabase(database);
	if (transaction != NULL) AttachTransaction(transaction);
}

BlobImpl::~BlobImpl()
{
	if (_Handle != NULL)
	{
		if (_WriteMode) Cancel();
		else try {Close();} catch (IBPP::Exception&) {}
	}
	if (_Transaction != NULL) _Transaction->DetachBlob(this);
	if (_Database != NULL) _Database->DetachBlob(this);
}

//
//	EOF
//
