///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: ibpp.h,v 1.27 2001/11/10 12:52:58 epocman Exp $
//	Subject : IBPP public header file. This is _the_ file you include in your
//			  application files when developing with IBPP.
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	Tabulations should be set every four characters when editing this file.
//
//	When compiling an IBPP project (or IBPP library itself), the following
//	defines should be made on the command-line (or in makefiles) according
//	to the OS platform and compiler used.
//
//	IBPP_WINDOWS or IBPP_LINUX			: to select the platform
//	IBPP_BCC or IBPP_GCC or IBPP_MSVC	: to select the compiler/environment
//
//	See the documentation and makefiles for more information.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __IBPP_H__
#define __IBPP_H__

#ifndef NULL
#define NULL	0
#endif

#if !defined(IBPP_WINDOWS) && !defined(IBPP_LINUX)
#error Please define IBPP_WINDOWS or IBPP_LINUX before compiling !
#endif

#if !defined(IBPP_BCC) && !defined(IBPP_GCC) && !defined(IBPP_MSVC)
#error Please define IBPP_BCC or IBPP_GCC or IBPP_MSVC before compiling !
#endif

#ifdef IBPP_WINDOWS
#include <windows.h>
#else
typedef long long __int64;
#endif

#if defined(IBPP_BCC) || defined(IBPP_MSVC)
#define HAS_HDRSTOP
#endif

namespace IBPP
{
	//	Typically you use this constant in a call IBPP::CheckVersion as in :
	//	if (! IBPP::CheckVersion(IBPP::Version)) { throw .... ; }
	const unsigned long Version = 0x01000502; // Version 01.00.05.02

	//	Transaction Access Modes
	typedef enum {amWrite, amRead} TAM;

	//	Transaction Isolation Levels
	typedef enum {ilConcurrency, ilReadDirty,
		ilReadCommitted, ilConsistency} TIL;

	//	Transaction Lock Resolution
	typedef enum {lrWait, lrNoWait} TLR;

	//	Prepared STatement Type
	typedef enum {stUnknown, stUnsupported,
		stSelect, stInsert, stUpdate, stDelete,	stDDL, stExecProcedure,
		stSelectUpdate, stOther} STT;

	//	SQL Data Types
	typedef enum {sdArray, sdBlob, sdDate, sdTime, sdTimestamp, sdString,
				sdSmallint, sdInteger, sdLargeint, sdFloat, sdDouble} SDT;

	//	Some forward declarations to keep the compiler happy
	class Database;
	class Transaction;
	class Statement;

	//	Class EventInterface is a just an interface.
	//	The Event handling system from the class Database requires the
	//	programmer to give an object pointer when defining a new event.
	//	When that event will fire, IBPP will call the method ibppEventHandler
	//	on the object registered with the event.
	//	So in order to catch events, the programmer must derive a class from
	//	EventInterface, implement 'ibppEventHandler' to do anything wanted
	//	when the event triggers and register an instance of that derived class
	//	when defining the event trap.

	class EventInterface
	{
	public:
		virtual void ibppEventHandler(Database*, const char*, int) = 0;

	protected:
		EventInterface() { };
		virtual ~EventInterface() { };
	};

	//	Classes Date, Time, Timestamp and DBKey are 'helper' classes.
	//	They are not strictly needed in application programs using IBPP. You
	//	can use the library to build Interbase 6 client programs without using
	//	them. Those classes merely represents specific data types and offer
	//	methods, constructors, copy constructors and so on adequate for classes
	//	meant to be used as data types. Their implementation details are not
	//	hidden (though private) from this file. You will discover from the
	//	comments in this file and the documentation that other classes of IBPP
	//	have been designed following a radically different approach.

	//	Class Date represent purely a Date (no time part specified).
	//	It is usefull in interactions with the SQL DATE type of Interbase.
	//	You can add or substract a number from a Date, that will modify it to
	//	represent the correct date, X days later or sooner. All the Y2K details
	//	taken into account. The assignment operator and conversion operator
	//	that deal with 'int' type are used to assign and extract the date as
	//	a stand alone integer number. This number represents the date as the
	//	number of days elapsed since 31 Dec 1899. That is 1 is 1 Jan 1900,
	//	2 is 2 Jan 1900 and so on... See the IBPP::dtoi and IBPP::itod methods.

	class Date
	{
	private:
		int _Date;		// The date as an integer : 1 == 1 Jan 1900

	public:
		void Clear(void) { _Date = 0; };
		void Today(void);
		void SetDate(int year, int month, int day);
		void GetDate(int& year, int& month, int& day);
		void Add(int days);
		void StartOfMonth(void);
		void EndOfMonth(void);
	
		Date() { Clear(); }
		Date(int year, int month, int day);
		Date(const Date&);					// Copy Constructor
		Date(const int&);
		Date& operator=(const Date&);		// Date Assignment operator
		Date& operator=(const int&);		// int Assignment operator
		operator int() { return _Date; }	// int Conversion operator
		~Date() { };
	};

	//	Class Time represent purely a Time (H:M:S).
	//	It is usefull in interactions with the SQL TIME type of Interbase.
	//	The assignment operator and conversion operator that deal with 'int'
	//	type are used to assign and extract the time as a stand alone integer
	//	number. This number represents the time as the number of seconds
	//	elapsed since midnight. See the IBPP::ttoi and IBPP::itot methods.

	class Time
	{
	private:
		int _Time;		// The time as an integer, seconds since midnight

	public:
		void Clear(void) { _Time = 0; }
		void Now(void);
		void SetTime(int hour, int minute, int second);
		void GetTime(int& hour, int& minute, int& second);

		Time() { Clear(); }
		Time(int hour, int minute, int second);
		Time(const Time&);					// Copy Constructor
		Time(const int&);
		Time& operator=(const Time&);		// Time Assignment operator
		Time& operator=(const int&);		// int Assignment operator
		operator int() { return _Time; }	// int Conversion operator
		~Time() { };
	};

	//	Class Timestamp represent a date AND a time.
	//	It is usefull in interactions with the SQL TIMESTAMP type of Interbase.

	class Timestamp
	{
	private:
		int _Date;		// The date as an integer : 1 == 1 Jan 1900
		int _Time;		// The time as an integer, seconds since midnight

	public:
		void Clear(void) { _Date = 0; _Time = 0; }
		void Today(void);
		void Now(void);

		void SetDate(int year, int month, int day);
		void GetDate(int& year, int& month, int& day);
		void SetTime(int hour, int minute, int second);
		void GetTime(int& hour, int& minute, int& second);
		void Add(int days);
		void StartOfMonth(void);
		void EndOfMonth(void);

		Timestamp() { Clear(); }
	  	Timestamp(int year, int month, int day);
		Timestamp(int year, int month, int day, int hour, int minute, int second);
		Timestamp(const Timestamp&);			// Copy Constructor
		Timestamp& operator=(const Timestamp&);	// Timestamp Assignment operator
		Timestamp& operator=(const Date&);		// Date Assignment operator
		Timestamp& operator=(const Time&);		// Time Assignment operator
		~Timestamp() { };
	};

	class DBKey
	{
	private:
		char* _DBKey;				// Stores the binary DBKey
		char* _String;				// String representation
		int _Size;					// Length of DBKey (multiple of 8 bytes) 

		void BlindCopy(const DBKey&);

	public:
		void SetKey(const char*, int size);
		void GetKey(char*, int size);
		const char* AsString(void);

		DBKey();
		DBKey(const DBKey&);			// Copy Constructor
		DBKey& operator=(const DBKey&);	// Assignment operator
		~DBKey();
	};

	//	Classes Blob, Service, Database, Transaction & Statement are abstract
	//	classes, with their constructors hidden. They can only be used through
	//	pointer access (->) and as they hide their constructors to enforce this
	//	behaviour, they can only be constructed by a helper method : a static
	//	class of the same name, ending in 'Factory'. For instance, to obtain a
	//	reference on a Blob object, you must call BlobFactory. In addition,
	//	the destructors are also hidden. You can't 'delete' those pointers.
	//	You have to call the method 'Release' on those objects to get them
	//	harakiri themselves. Those objects follow this convention for reasons
	//	too long to describe here. See the documentation as soon as it will be
	//	written. Such objects have all the methods needed to be re-useable.
	//	One Database object can be used to connect a database, and when that
	//	connection will get closed later, it is possible to reuse the same
	//	object to connect to another database. The equivalent behaviour stands
	//	true for Transaction, Statement, Service, and Blob.
	//	Think of those classes and objects as the building blocks of the
	//	procedural body of the application program. This design as been chosen
	//	to better control leakage and bad pointers useage on those basic
	//	building blocks. Classes DBKey, Date, Time and Timestamp above are
	//	just 'helper' classes representing data types. That's why they were
	//	designed in a more classical way.

	//	Class Exception can't be instantiated, just as discussed above.
	//	In addition this class does not even have a 'Factory'.
	//	The conclusion is this : there is NO way you can construct an object
	//	of this type by yourself. Its only purpose is to catch C++ exceptions
	//	thrown from the IBPP implementation itself. To achieve this, IBPP
	//	throws exceptions through a derived class of 'Exception'. By virtue
	//	of C++ polymorphism, you can catch those as references to Exception as
	//	in : .... catch (Exception& e) { e.PrintfErrorMessage(); }

	class Exception
	{
	public:
		virtual const char* GetOrigin(void) = 0;
		virtual void GetErrorMessage(char*) = 0;
		virtual int SqlCode(void) = 0;
		virtual void PrintfErrorMessage(void) = 0;

#ifndef _MSC_VER  // MSVC++ bug workaround
	protected:
#endif
	    Exception() { };
		virtual ~Exception() { };
	};

	//	Class Blob adopts the full behaviour described above.
	//	You need to use IBPP::BlobFactory to get a pointer to an instance.
	//	You need to call ->Release() on that instance to free it.
	//	With a Blob object, you can create, read and write Interbase blobs.

	class Blob
	{
	public:
		virtual void AttachDatabase(Database*) = 0;
		virtual void AttachTransaction(Transaction*) = 0;
		virtual void DetachDatabase(void) = 0;
		virtual void DetachTransaction(void) = 0;
		virtual Database* GetDatabase(void) = 0;
		virtual Transaction* GetTransaction(void) = 0;
		virtual void Create(void) = 0;
		virtual void Open(void) = 0;
		virtual void Close(void) = 0;
		virtual void Cancel(void) = 0;
		virtual int Read(void*, int size) = 0;
		virtual void Write(void*, int size) = 0;
		virtual void GetInfo(int* Size, int* Largest, int* Segments) = 0;

		virtual void Release(Blob*&) = 0;

	protected:
		Blob() { };
		virtual ~Blob() { };
	};

	//	Class Array adopts the full behaviour described above.
	//	You need to use IBPP::ArrayFactory to get a pointer to an instance.
	//	You need to call ->Release() on that instance to free it.
	//	With an Array object, you can create, read and write Interbase Arrays.

	class Array
	{
	public:
		virtual void AttachDatabase(Database*) = 0;
		virtual void AttachTransaction(Transaction*) = 0;
		virtual void DetachDatabase(void) = 0;
		virtual void DetachTransaction(void) = 0;
		virtual Database* GetDatabase(void) = 0;
		virtual Transaction* GetTransaction(void) = 0;
		/*
		virtual void Create(void) = 0;
		virtual void Open(void) = 0;
		virtual void Close(void) = 0;
		virtual void Cancel(void) = 0;
		virtual int Read(void*, int size) = 0;
		virtual void Write(void*, int size) = 0;
		virtual void GetInfo(int* Size, int* Largest, int* Segments) = 0;
		*/

		virtual void Release(Array*&) = 0;

	protected:
		Array() { };
		virtual ~Array() { };
	};

	//	Class Service adopts the full behaviour described above.
	//	You need to use IBPP::ServiceFactory to get a pointer to an instance.
	//	You need to call ->Release() on that instance to free it.
	//	With a Service object, you can do some maintenance work of databases
	//	and servers (backup, restore, create/update users).

	class Service
	{
	public:
	    virtual void Connect(void) = 0;
		virtual void Disconnect(void) = 0;

		virtual void AddUser(const char* username, const char* password,
			const char* first = NULL, const char* middle = NULL, const char* last = NULL) = 0;
		virtual void ModifyUser(const char* username, const char* password = NULL,
			const char* first = NULL, const char* middle = NULL, const char* last = NULL) = 0;
		virtual void RemoveUser(const char* username) = 0;

		virtual void StartBackup(const char* dbfile, const char* bkfile,
			bool verbose = false) = 0;
		virtual void StartRestore(const char* bkfile, const char* dbfile,
			int pagesize = 0, bool verbose = false) = 0;

		virtual bool Wait(char* line) = 0;	// With reporting (does not block)
		virtual bool Wait(void) = 0;		// Without reporting (does block)

		virtual void Release(Service*&) = 0;

	protected:
	    Service() { };
		virtual ~Service() { };
	};

	//	Class Database adopts the full behaviour described above.
	//	You need to use IBPP::DatabaseFactory to get a pointer to an instance.
	//	You need to call ->Release() on that instance to free it.
	//	With a Database object, you can create/drop/connect databases.

	class Database
	{
	public:
		virtual const char* GetServerName(void) = 0;
		virtual void SetServerName(const char* ServerName) = 0;
		virtual const char* GetDatabaseName(void) = 0;
		virtual void SetDatabaseName(const char* DatabaseName) = 0;
		virtual const char* GetUsername(void) = 0;
		virtual void SetUsername(const char* UserName) = 0;
		virtual const char* GetUserPassword(void) = 0;
		virtual void SetUserPassword(const char* UserPassword) = 0;
		virtual const char* GetRoleName(void) = 0;
		virtual void SetRoleName(const char* RoleName) = 0;
		virtual const char* GetCreateParams(void) = 0;
		virtual void SetCreateParams(const char* CreateParams) = 0;
		virtual void AttachTransaction(Transaction*) = 0;
		virtual void DetachTransaction(Transaction*) = 0;
		virtual void AttachStatement(Statement*) = 0;
		virtual void DetachStatement(Statement*) = 0;
		virtual void AttachBlob(Blob*) = 0;
		virtual void DetachBlob(Blob*) = 0;

		virtual void GetInfo(int* ODS, int* ODSMinor, int* PageSize, bool* Sync) = 0;
		virtual void GetStat(int* Pages, int* Memory, int* Buffers,
			int* Fetches, int* marks, int* reads, int* writes) = 0;

	    virtual void Create(void) = 0;
		virtual void Connect(void) = 0;
		virtual void Disconnect(void) = 0;
	    virtual void Drop(void) = 0;

		virtual void DefineEvent(const char*, EventInterface*) = 0;
		virtual void DropEvent(const char*) = 0;

		virtual void ClearEvents(void) = 0;
		virtual void DispatchEvents(void) = 0;

		virtual void Release(Database*&) = 0;

	protected:
	    Database() { };
	    virtual ~Database() { };
	};

	//	Class Transaction adopts the full behaviour described above.
	//	You need to use IBPP::TransactionFactory to get a pointer to an instance.
	//	You need to call ->Release() on that instance to free it.
	//	A Transaction object can be associated with more than one Database,
	//	allowing for distributed transactions spanning multiple databases,
	//	possibly located on different servers.

	class Transaction
	{
	public:
	    virtual void AttachDatabase(Database* db, TAM am = amWrite,
				TIL il = ilConcurrency, TLR lr = lrWait) = 0;
	    virtual void DetachDatabase(Database* db) = 0;
	    virtual void Start(void) = 0;
	    virtual void Commit(void) = 0;
	    virtual void Rollback(void) = 0;
	    virtual void CommitRetain(void) = 0;
		virtual void RollbackRetain(void) = 0;

		virtual void Release(Transaction*&) = 0;

	protected:
		Transaction() { };
	    virtual ~Transaction() { };
	};

	//	Class Statement adopts the full behaviour described above.
	//	You need to use IBPP::StatementFactory to get a pointer to an instance.
	//	You need to call ->Release() on that instance to free it.
	//	A Statement object is the work horse of IBPP. All your data
	//	manipulation statements will be done through it. It is also used to
	//	access the result set of a query (when the statement is such).
	//	Today, there is no independent 'ResultSet' object per-se in IBPP.

	class Statement
	{
	public:
		virtual void AttachDatabase(Database*) = 0;
		virtual void DetachDatabase(void) = 0;
		virtual void AttachTransaction(Transaction*) = 0;
		virtual void DetachTransaction(void) = 0;
		virtual void Prepare(const char*) = 0;
		virtual void Execute(const char* = NULL) = 0;
		virtual void ExecuteImmediate(const char*) = 0;
		virtual void CursorExecute(const char* cursor, const char* sql = NULL) = 0;
		virtual bool Fetch(void) = 0;
		virtual int AffectedRows(void) = 0;
		virtual void Close(void) = 0;
		virtual STT GetType(void) = 0;

		virtual void SetNull(int) = 0;
		virtual void Set(int, bool) = 0;
		virtual void Set(int, const char*) = 0;
		virtual void Set(int, short value) = 0;
		virtual void Set(int, int value) = 0;
		virtual void Set(int, long value) = 0;
		virtual void Set(int, __int64 value) = 0;
		virtual void Set(int, float value) = 0;
		virtual void Set(int, double value) = 0;
		virtual void Set(int, Timestamp& value) = 0;
		virtual void Set(int, Date& value) = 0;
		virtual void Set(int, Time& value) = 0;
		virtual void Set(int, DBKey& value) = 0;
		virtual void Set(int, Blob* value) = 0;
		virtual void Set(int, Array* value) = 0;

		virtual bool IsNull(int) = 0;
		virtual bool Get(int, bool*) = 0;
		virtual bool Get(int, char*) = 0;
		virtual bool Get(int, short*) = 0;
		virtual bool Get(int, int*) = 0;
		virtual bool Get(int, long*) = 0;
		virtual bool Get(int, __int64*) = 0;
		virtual bool Get(int, float*) = 0;
		virtual bool Get(int, double*) = 0;
		virtual bool Get(int, Timestamp& value) = 0;
		virtual bool Get(int, Date& value) = 0;
		virtual bool Get(int, Time& value) = 0;
		virtual bool Get(int, DBKey& value) = 0;
		virtual bool Get(int, Blob* value) = 0;
		virtual bool Get(int, Array* value) = 0;

		virtual bool IsNull(const char*) = 0;
		virtual bool Get(const char*, bool*) = 0;
		virtual bool Get(const char*, char*) = 0;
		virtual bool Get(const char*, short*) = 0;
		virtual bool Get(const char*, int*) = 0;
		virtual bool Get(const char*, long*) = 0;
		virtual bool Get(const char*, __int64*) = 0;
		virtual bool Get(const char*, float*) = 0;
		virtual bool Get(const char*, double*) = 0;
		virtual bool Get(const char*, Timestamp& value) = 0;
		virtual bool Get(const char*, Date& value) = 0;
		virtual bool Get(const char*, Time& value) = 0;
		virtual bool Get(const char*, DBKey& value) = 0;
		virtual bool Get(const char*, Blob* value) = 0;
		virtual bool Get(const char*, Array* value) = 0;

		virtual int ColumnNum(const char*) = 0;
		virtual const char* ColumnName(int) = 0;
		virtual const char* ColumnAlias(int) = 0;
		virtual const char* ColumnTable(int) = 0;
		virtual SDT ColumnType(int) = 0;
		virtual int ColumnSize(int) = 0;
		virtual int ColumnScale(int) = 0;

		virtual int Columns(void) = 0;

		virtual void Release(Statement*&) = 0;

	protected:
		Statement() { };
	    virtual ~Statement() { };
	};

	//	Factories for the above objects. These methods are the only way
	//	to instantiate and get a pointer to a operative object like Database,
	//	Statement, and so on...

	Service* ServiceFactory(const char* ServerName,
		const char* UserName, const char* UserPassword);

	Database* DatabaseFactory(const char* ServerName = NULL,
		const char* DatabaseName = NULL, const char* UserName = NULL,
			const char* UserPassword = NULL, const char* RoleName = NULL);

	Transaction* TransactionFactory(Database* db,
		TAM am = amWrite, TIL il = ilConcurrency, TLR lr = lrWait);

	Statement* StatementFactory(Database* db, Transaction* tr,
		const char* sql = NULL);

	Blob* BlobFactory(Database* db, Transaction* tr);

	Array* ArrayFactory(Database* db, Transaction* tr);

	//	IBPP uses a self initialization system. Each time an object that may
	//	require the usage of the Interbase client C-API library is used, the
	//	library internal handling details are automatically initialized, if not
	//	already done. You can kick this initialization at the start of an
	//	application by calling IBPP::CheckVersion(). This is recommended,
	//	because IBPP::CheckVersion will assure you that YOUR code has been
	//	compiled against a compatible version of the library.
	//	Ideally, you would call IBPP::Release() at the end of your application
	//	program to free every resource used by the library. Forgetting to do
	//	this is often not harmfull, though.

	bool CheckVersion(unsigned long);
	int GDSVersion(void);
	void Release(void);

	//	Finally, here are some date and time conversion routines used by IBPP
	//	and that may be helpful at the application level. They do not depend
	//	on anything related to Interbase.

	void dtoi (int date, int *y, int *m, int *d);
	void itod (int* pdate, int year, int month, int day);
	void ttoi (int itime, int *hour, int *minute, int *second);
	void itot (int* ptime, int hour, int minute, int second);

};

#endif

//
//	EOF
//

